@@ -1,4 +1,4 @@
-# Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES
+# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES
 #
 # SPDX-License-Identifier: BSD-3-Clause
 
@@ -10,7 +10,7 @@ if not jax.config.jax_enable_x64:
     raise RuntimeError(f"jax_enable_x64 must be set to True to use cuQuantum Python JAX")
 
 from .operator_action import operator_action
-from .pysrc.operators import (
+from .pysrc import (
     ElementaryOperator,
     MatrixOperator,
     OperatorTerm,
